import json
import sys
from inspect import signature

from bitstring import BitArray, BitStream

from hivemind_bus_client.exceptions import UnsupportedProtocolVersion
from hivemind_bus_client.message import HiveMessageType, HiveMessage, HiveMindBinaryPayloadType
from hivemind_bus_client.util import compress_payload, decompress_payload, cast2bytes, bytes2str

PROTOCOL_VERSION = 1  # integer, a version increase signals new functionality added
                      # version 0 is the original hivemind protocol, 1 supports handshake + binary


_INT2TYPE = {0: HiveMessageType.HANDSHAKE,
             1: HiveMessageType.BUS,
             2: HiveMessageType.SHARED_BUS,
             3: HiveMessageType.BROADCAST,
             4: HiveMessageType.PROPAGATE,
             5: HiveMessageType.ESCALATE,
             6: HiveMessageType.HELLO,
             7: HiveMessageType.QUERY,
             8: HiveMessageType.CASCADE,
             9: HiveMessageType.PING,
             10: HiveMessageType.RENDEZVOUS,
             11: HiveMessageType.THIRDPRTY,
             12: HiveMessageType.BINARY}


def get_bitstring(hive_type=HiveMessageType.BUS, payload=None,
                  compressed=None, hivemeta=None,
                  binary_type=HiveMindBinaryPayloadType.UNDEFINED,
                  proto_version=PROTOCOL_VERSION, versioned=False):
    if proto_version <= 1:
        if compressed is None:  # auto
            unc = _get_bitstring_v1(hive_type, payload, False, hivemeta, binary_type, versioned)
            comp = _get_bitstring_v1(hive_type, payload, True, hivemeta, binary_type, versioned)
            if len(unc) <= len(comp):
                return unc
            return comp
        return _get_bitstring_v1(hive_type, payload, bool(compressed), hivemeta, binary_type, versioned)
    raise UnsupportedProtocolVersion(f"Max Supported Version: {PROTOCOL_VERSION}")


def _get_bitstring_v1(hive_type=HiveMessageType.BUS, payload=None,
                      compressed=True, hivemeta=None,
                      binary_type=HiveMindBinaryPayloadType.UNDEFINED, versioned=False):
    # there are 12 hivemind message main types
    typemap = {v: k for k, v in _INT2TYPE.items()}
    binmap = {e: e.value for e in HiveMindBinaryPayloadType}

    s = BitArray()
    s.append(f'uint:1={int(1)}')  # always start with a 1, 0s to the left for padding so it can be cast to bytes
    s.append(f'uint:1={int(versioned)}')  # 1 bit unsigned integer - requires protocol version
    if versioned:
        s.append(f'uint:8={PROTOCOL_VERSION}')
    s.append(f'uint:5={typemap.get(hive_type, 11)}')  # 5 bit unsigned integer - the hive msg type
    s.append(f'uint:1={int(bool(compressed))}')  # 1 bit unsigned integer - payload is zlib compressed

    # NOTE: hivemind meta is reserved TBD arbitrary data
    hivemeta = cast2bytes(hivemeta or {}, compressed)
    s.append(f'uint:8={len(hivemeta)}')  # 8 bit unsigned integer - N of bytes for metadata
    s.append(hivemeta)  # arbitrary hivemind meta

    # when payload is binary data meant to be passed along raw and not parsed
    if hive_type == HiveMessageType.BINARY:
        # 4 bit unsigned integer - integer indicating pseudo format of bin content
        s.append(f'uint:4={binmap.get(binary_type, 0)}')
    # the remaining bits are the payload
    else:
        if hasattr(payload, "serialize"):
            payload = payload.serialize()
        payload = cast2bytes(payload, compressed)

    s.append(payload)

    # pad
    while len(s) % 8 != 0:
        s.insert(f'uint:1={int(0)}', 0)

    return s


def decode_bitstring(bitstr):
    s = BitStream(bitstr)
    pad = False
    while not pad:
        pad = s.read(1).bool
    versioned = s.read(1).bool
    if versioned:
        proto_version = s.read(8).uint
    else:
        proto_version = PROTOCOL_VERSION
    if proto_version <= 1:
        return _decode_bitstring_v1(s)
    raise UnsupportedProtocolVersion(f"Max Supported Version: {PROTOCOL_VERSION}")


def _decode_bitstring_v1(s):
    binmap = {e: e.value for e in HiveMindBinaryPayloadType}

    hive_type = _INT2TYPE.get(s.read(5).uint, 11)
    compressed = bool(s.read(1))

    metalen = s.read(8).uint * 8
    meta = s.read(metalen)

    # TODO standardize hivemind meta
    meta = json.loads(bytes2str(meta.bytes, compressed))

    is_bin = hive_type == HiveMessageType.BINARY
    bin_type = HiveMindBinaryPayloadType.UNDEFINED
    if is_bin:
        bin_type = binmap.get(s.read(4).uint, 0)

    payload_len = len(s) - s.pos
    payload = s.read(payload_len)

    if not is_bin:
        payload = bytes2str(payload.bytes, compressed)
    else:
        payload = payload.bytes

    return HiveMessage(hive_type, payload,
                       metadata=meta, bin_type=bin_type)


def mycroft2bitstring(msg, compressed=False):
    if isinstance(msg, str):
        msg = Message.deserialize(msg)
    return get_bitstring(HiveMessageType.BUS, payload=msg, hivemeta=msg.context, compressed=compressed)


if __name__ == "__main__":
    d = {e: e.value for e in HiveMindBinaryPayloadType}
    from hivemind_bus_client.message import Message

    text = """The Mycroft project is also working on and selling smart speakers that run its software. All of its hardware is open-source, released under the CERN Open Hardware Licence.
Its first hardware project was the Mark I, targeted primarily at developers. Its production was partially funded through a Kickstarter campaign, which finished successfully. Units started shipping out in April 2016.
Its most recent hardware project is the Mark II, intended for general usage, not just for developers. Unlike the Mark I, the Mark II is equipped with a screen, being able to relay information both visually as well as acoustically. As with the Mark I, the Mark II's production was partially funded through a Kickstarter campaign, which wrapped up in February 2018, hitting almost 8 times its original goal. As of February 2021, the Mark II had not yet begun shipping to crowd-funders, though shipping of the Development Kit was imminent.
Mycroft announced that a third hardware project, Mark III, will be offered through Kickstarter, and that an entire product line of Mark I, II, and III will be released to stores by November, 2019"""

    payload = Message("speak", {"utterance": text})

    json_plod = HiveMessage(HiveMessageType.BUS, payload)
    n_json_bits = len(json_plod.serialize().encode("utf-8")) * 8
    # 10096 - naive json2bytes

    bitstr = get_bitstring(hive_type=HiveMessageType.BUS,
                           payload=payload,
                           compressed=False)
    n_unc_bits = len(bitstr)
    # 9494  - uncompressed HM

    bitstr = get_bitstring(hive_type=HiveMessageType.BUS,
                           payload=payload,
                           compressed=True)
    n_enc_bits = len(bitstr)
    # 4886 - compressed HM  (small strings might actually become larger)

    decoded = decode_bitstring(bitstr)
    print(decoded)

    payload = HiveMessage(HiveMessageType.BUS,
                          payload=Message("speak", {"utterance": "RED ALERT"}))
    bitstr = get_bitstring(hive_type=HiveMessageType.BROADCAST,
                           payload=payload,
                           compressed=False)

    decoded = decode_bitstring(bitstr)
    print(decoded)

    compressed = compress_payload(text).hex()


    # 789c5590c16e84300c44ef7cc51c5ba942bdee1fb4528ffb03261848156c9418e8fe7d9daebab0b72863cfbcf1758a05c32ac1a20afc6d1363c971a67c4314e33c506098bae0eaacfd9a18945446ecd126343d079d97cca5bcbc3e9c5a5c9f8c33db9aa5a0bb1943bb6f0ee66ffc6f4677abc13d19a119e3c65223a3810a16ca34b39354533e3c27d748d4f7f231834029718fc41b27ec530c8e18542c6bba97e31f6331e870a457de4fcf92bfc6a3bb746c3b3bc47bc5b8b4f8d29d8b61a3b4b27f36c5487aefa719a2672337e971c149efeae253d4471c2b7385b9633a4b739a78c39899ec3122a3dff9c4ebf54e776c7f0106a5a377

    def measure_compression(text):
        text = text.encode("utf-8")
        text_size = sys.getsizeof(text)
        print("N bytes of original text", text_size)

        compressed = compress_payload(text)
        csize = sys.getsizeof(compressed)
        print("N bytes of compressed text", csize)

        decompressed = decompress_payload(compressed)
        dsize = sys.getsizeof(decompressed)
        print("N bytes of decompressed text", dsize)

        sdiff = text_size - csize
        print("Difference of N bytes", sdiff)

        print("N bytes reduced by", sdiff * 100 / text_size, "%")

        return sdiff * 100 / text_size


    measure_compression(text)
    # N bytes of original text 1153
    # N bytes of compressed text 588
    # N bytes of decompressed text 1153
    # Difference of N bytes 565
    # N bytes reduced by 49.00260190806591 %



import json
import zlib
from typing import Union, Dict

from hivemind_bus_client.encryption import SupportedEncodings, SupportedCiphers
from hivemind_bus_client.message import HiveMessage, HiveMessageType, Message


def serialize_message(message: Union[HiveMessage, Message, Dict]) -> str:
    # convert a Message object into raw data that can be sent over
    # websocket
    if hasattr(message, 'serialize'):
        return message.serialize()
    elif isinstance(message, dict):
        message = {
            k: v if not hasattr(v, 'serialize') else serialize_message(v)
            for k, v in message.items()}
        return json.dumps(message)
    else:
        return json.dumps(message.__dict__)


def payload2dict(payload: Union[HiveMessage, Message, str]) -> Dict:
    """helper to ensure all subobjects of a payload are a dict safe for serialization
    eg. ensure payload is valid to send over mycroft messagebus object """
    if isinstance(payload, HiveMessage):
        payload = payload.as_dict
    if isinstance(payload, Message):
        payload = payload.serialize()
    if isinstance(payload, str):
        try:
            payload = json.loads(payload)
        except:
            pass
    assert isinstance(payload, dict)

    def can_serialize(val):
        if isinstance(val, HiveMessage) \
                or isinstance(val, Message) \
                or isinstance(val, dict):
            return True
        return False

    for k, v in payload.items():
        if can_serialize(v):
            payload[k] = payload2dict(v)
        if isinstance(v, list):
            for idx, l in enumerate(v):
                if can_serialize(l):
                    payload[k][idx] = payload2dict(l)
    return payload


def get_payload(msg: Union[HiveMessage, Message, str, Dict]) -> Dict:
    """ helper to read normalized payload
    from all supported formats (HiveMessage, Message, json str)
    """
    if isinstance(msg, HiveMessage):
        msg = msg.payload
    if isinstance(msg, Message):
        msg = msg.serialize()
    if isinstance(msg, str):
        msg = json.loads(msg)
    return msg


def get_hivemsg(msg: Union[Message, str, Dict]) -> HiveMessage:
    """ helper to create a normalized HiveMessage object
    from all supported formats (Message, json str, dict)
    """
    if isinstance(msg, str):
        msg = json.loads(msg)
    if isinstance(msg, dict):
        msg = HiveMessage(**msg)
    if isinstance(msg, Message):
        msg = HiveMessage(msg_type=HiveMessageType.BUS, payload=msg)
    assert isinstance(msg, HiveMessage)
    return msg


def get_mycroft_msg(pload: Union[HiveMessage, str, Dict]) -> Message:
    if isinstance(pload, HiveMessage):
        assert pload.msg_type == HiveMessageType.BUS
        pload = pload.payload

    if isinstance(pload, str):
        try:
            pload = Message.deserialize(pload)
        except:
            pload = json.loads(pload)
    if isinstance(pload, dict):
        msg_type = pload.get("msg_type") or pload["type"]
        data = pload.get("data") or {}
        context = pload.get("context") or {}
        pload = Message(msg_type, data, context)

    assert isinstance(pload, Message)
    return pload


def compress_payload(text: Union[str, bytes]) -> bytes:
    # Compressing text
    if isinstance(text, str):
        decompressed = text.encode("utf-8")
    else:
        decompressed = text
    return zlib.compress(decompressed)


def decompress_payload(compressed: bytes) -> bytes:
    return zlib.decompress(compressed)


def cast2bytes(payload: Union[Dict, str], compressed=False) -> bytes:
    if isinstance(payload, dict):
        payload = json.dumps(payload)
    if compressed:
        payload = compress_payload(payload)
    if isinstance(payload, str):
        payload = payload.encode("utf-8")
    assert isinstance(payload, bytes)
    return payload


def bytes2str(payload: bytes, compressed=False) -> str:
    if compressed:
        return decompress_payload(payload).decode("utf-8")
    else:
        return payload.decode("utf-8")


###############
# deprecated
import warnings


def encrypt_as_json(key, data, b64=False) -> str:
    warnings.warn(
        "encrypt_as_json is deprecated and will be removed in future versions. "
        "Use 'from hivemind_bus_client.encryption import encrypt_as_json' instead",
        DeprecationWarning,
        stacklevel=2
    )
    from hivemind_bus_client.encryption import encrypt_as_json as _ej
    c = SupportedEncodings.JSON_B64 if b64 else SupportedEncodings.JSON_HEX
    return _ej(key, data, encoding=c, cipher=SupportedCiphers.AES_GCM)


def decrypt_from_json(key, data: Union[str, bytes]):
    warnings.warn(
        "decrypt_from_json is deprecated and will be removed in future versions. "
        "Use 'from hivemind_bus_client.encryption import decrypt_from_json' instead",
        DeprecationWarning,
        stacklevel=2
    )
    from hivemind_bus_client.encryption import decrypt_from_json as _dj
    try:
        return _dj(key, data, encoding=SupportedEncodings.JSON_HEX, cipher=SupportedCiphers.AES_GCM)
    except Exception as e:
        try:
            return _dj(key, data, encoding=SupportedEncodings.JSON_B64, cipher=SupportedCiphers.AES_GCM)
        except:
            raise e


def encrypt_bin(key, data: Union[str, bytes]):
    warnings.warn(
        "encrypt_bin is deprecated and will be removed in future versions. "
        "Use 'from hivemind_bus_client.encryption import encrypt_bin' instead",
        DeprecationWarning,
        stacklevel=2
    )
    from hivemind_bus_client.encryption import encrypt_bin as _eb
    return _eb(key, data, cipher=SupportedCiphers.AES_GCM)


def decrypt_bin(key, ciphertext: bytes):
    warnings.warn(
        "decrypt_bin is deprecated and will be removed in future versions. "
        "Use 'from hivemind_bus_client.encryption import decrypt_bin' instead",
        DeprecationWarning,
        stacklevel=2
    )
    from hivemind_bus_client.encryption import decrypt_bin as _db
    return _db(key, ciphertext, SupportedCiphers.AES_GCM)


if __name__ == "__main__":
    k = "*" * 16
    test = "this is a test text for checking size of encryption and stuff"
    print(len(test))  # 61

    encjson = encrypt_as_json(k, test, b64=True)
    # {"ciphertext": "MkTc1LSK3jugt5SXapAeSrD6YWnYdSJ5oqF2bWYcnFpAYgjAgcTFXiKL3wBsqVKY52SkO5mjkqr7i/0A5A==", "tag": "37WNN8e23Mj0LlOxu9cjnQ==", "nonce": "inRwcb0H1Xu6pp80WFeJvg=="}
    print(len(encjson))  # 174
    assert decrypt_from_json(k, encjson) == test

    encjson = encrypt_as_json(k, test, b64=False)
    # {"ciphertext": "64c65bad86a3582097aa4958b7c9555e8bf7deeac6bdf8b5f648cc360aaf50062ae9c635f602b3c66b2de1eece57666b3412a26f55bbd5ace2f601d8c2", "tag": "ce550c1e399c92bb26bf3c171c212e7d", "nonce": "84d045071b05bf005145ce071df0ed41"}
    print(len(encjson))  # 228
    assert decrypt_from_json(k, encjson) == test



import enum
import json
from binascii import hexlify, unhexlify
from typing import Union, Optional, Dict, Any, Literal, List, Callable

import pybase64
from hivemind_bus_client.z85b import Z85B
from Cryptodome.Cipher import AES, ChaCha20_Poly1305
from cpuinfo import get_cpu_info

from hivemind_bus_client.exceptions import EncryptionKeyError, DecryptionKeyError, InvalidEncoding, InvalidCipher, \
    InvalidKeySize

# Cipher-specific constants
AES_KEY_SIZES = [16, 24, 32]  # poorman_handshake generates 32 bit secrets
AES_NONCE_SIZE = 16
AES_TAG_SIZE = 16
CHACHA20_KEY_SIZE = 32
CHACHA20_NONCE_SIZE = 12
CHACHA20_TAG_SIZE = 16


def cpu_supports_AES() -> bool:
    """
    Check if the CPU supports AES encryption.

    This function checks the CPU flags to determine if the hardware supports
    AES encryption. It does so by querying the CPU information and checking
    if the 'aes' flag is present.

    Returns:
        bool: True if AES is supported by the CPU, False otherwise.
    """
    return "aes" in get_cpu_info()["flags"]


class SupportedEncodings(str, enum.Enum):
    """
    Enum representing JSON-based encryption encodings.

    Ciphers output binary data, and JSON needs to transmit that data as plaintext.
    The supported encodings include Base64 and Hex encoding.
    """
    JSON_Z85B = "JSON-Z85B"  # JSON text output with Z85B encoding
    JSON_B64 = "JSON-B64"  # JSON text output with Base64 encoding
    JSON_HEX = "JSON-HEX"  # JSON text output with Hex encoding


def get_encoder(encoding: SupportedEncodings) -> Callable[[bytes], bytes]:
    encoding = _norm_encoding(encoding)
    if encoding == SupportedEncodings.JSON_B64:
        return pybase64.b64encode
    if encoding == SupportedEncodings.JSON_HEX:
        return hexlify
    if encoding == SupportedEncodings.JSON_Z85B:
        return Z85B.encode
    raise InvalidEncoding(f"Invalid encoding: {encoding}")


def get_decoder(encoding: SupportedEncodings) -> Callable[[bytes], bytes]:
    encoding = _norm_encoding(encoding)
    if encoding == SupportedEncodings.JSON_B64:
        return pybase64.b64decode
    if encoding == SupportedEncodings.JSON_HEX:
        return unhexlify
    if encoding == SupportedEncodings.JSON_Z85B:
        return Z85B.decode
    raise InvalidEncoding(f"Invalid encoding: {encoding}")


class SupportedCiphers(str, enum.Enum):
    """
    Enum representing binary encryption ciphers.

    Specifications:
      - AES - http://csrc.nist.gov/publications/fips/fips197/fips-197.pdf
      - GCM - http://csrc.nist.gov/publications/nistpubs/800-38D/SP-800-38D.pdf
      - CHACHA20-POLY1305 - https://datatracker.ietf.org/doc/html/rfc7539
    """
    AES_GCM = "AES-GCM"
    CHACHA20_POLY1305 = "CHACHA20-POLY1305"  # specified in RFC7539.


AES_CIPHERS = {c for c in SupportedCiphers if "AES" in c}
BLOCK_CIPHERS = AES_CIPHERS  # Blowfish etc can be added in the future


def optimal_ciphers() -> List[SupportedCiphers]:
    """
    Determine the optimal ciphers based on CPU support.

    This function checks if the CPU supports AES encryption. If it does, it
    returns a list of ciphers with AES first, followed by other supported ciphers.
    If AES is not supported, the function returns a list of ciphers with
    ChaCha20-Poly1305 first.

    Returns:
        List[SupportedCiphers]: A list of optimal ciphers based on CPU support.
    """
    if not cpu_supports_AES():
        return [SupportedCiphers.CHACHA20_POLY1305, SupportedCiphers.AES_GCM]
    return [SupportedCiphers.AES_GCM, SupportedCiphers.CHACHA20_POLY1305]


def _norm_cipher(cipher: Union[SupportedCiphers, str]) -> SupportedCiphers:
    """
    Normalize a cipher to an enum member.

    This function takes either a cipher string or an enum member and ensures it
    is converted to the corresponding enum member of SupportedCiphers. If the input
    is invalid, an InvalidCipher exception is raised.

    Args:
        cipher (Union[SupportedCiphers, str]): The cipher to normalize, either a string or an enum member.

    Returns:
        SupportedCiphers: The corresponding enum member of SupportedCiphers.

    Raises:
        InvalidCipher: If the cipher is invalid.
    """
    if isinstance(cipher, SupportedCiphers):
        return cipher  # If already an enum member, just return it

    # Convert string to enum member by matching the value
    for member in SupportedCiphers:
        if member.value == cipher:
            return member

    raise InvalidCipher(f"Invalid cipher: {cipher}")


def _norm_encoding(encoding: Union[SupportedEncodings, str]) -> SupportedEncodings:
    """
    Normalize an encoding to an enum member.

    This function takes either an encoding string or an enum member and ensures it
    is converted to the corresponding enum member of SupportedEncodings. If the input
    is invalid, an InvalidEncoding exception is raised.

    Args:
        encoding (Union[SupportedEncodings, str]): The encoding to normalize, either a string or an enum member.

    Returns:
        SupportedEncodings: The corresponding enum member of SupportedEncodings.

    Raises:
        InvalidEncoding: If the encoding is invalid.
    """
    if isinstance(encoding, SupportedEncodings):
        return encoding  # If already an enum member, just return it

    # Convert string to enum member by matching the value
    for member in SupportedEncodings:
        if member.value == encoding:
            return member

    raise InvalidEncoding(f"Invalid JSON encoding: {encoding}")


def encrypt_as_json(
        key: Union[str, bytes],
        plaintext: Union[str, Dict[str, Any]],
        cipher: Union[SupportedCiphers, str] = SupportedCiphers.AES_GCM,
        encoding: Union[SupportedEncodings, str] = SupportedEncodings.JSON_B64
) -> str:
    """
    Encrypts the given data and outputs it as a JSON string.

    Args:
        key (Union[str, bytes]): The encryption key, up to 16 bytes. Longer keys will be truncated.
        plaintext (Union[str, Dict[str, Any]]): The data to encrypt. If a dictionary, it will be serialized to JSON.
        cipher (Union[SupportedCiphers, str]): The encryption cipher. Supported options:
            - AES-GCM (default)
        encoding (Union[SupportedEncodings, str]): The encoding type for JSON. Supported options:
            - JSON-B64 (default)

    Returns:
        str: A JSON string containing the encrypted data, nonce, and tag.

    Raises:
        InvalidCipher: If an unsupported cipher is provided.
        InvalidEncoding: If an unsupported encoding is provided.
    """

    cipher = _norm_cipher(cipher)
    encoding = _norm_encoding(encoding)

    # If plaintext is a dictionary, convert it to a JSON string
    if isinstance(plaintext, dict):
        plaintext = json.dumps(plaintext)

    try:
        ciphertext = encrypt_bin(key=key, plaintext=plaintext, cipher=cipher)
    except InvalidKeySize as e:
        raise e
    except Exception as e:
        raise EncryptionKeyError from e

    # Extract nonce/tag depending on cipher, sizes are different
    if cipher in AES_CIPHERS:
        nonce, ciphertext, tag = (
            ciphertext[:AES_NONCE_SIZE],
            ciphertext[AES_NONCE_SIZE:-AES_TAG_SIZE],
            ciphertext[-AES_TAG_SIZE:]
        )
    else:
        nonce, ciphertext, tag = (
            ciphertext[:CHACHA20_NONCE_SIZE],
            ciphertext[CHACHA20_NONCE_SIZE:-CHACHA20_TAG_SIZE],
            ciphertext[-CHACHA20_TAG_SIZE:]
        )

    # Choose encoder based on the encoding
    encoder = get_encoder(encoding)

    # Return the JSON-encoded ciphertext, tag, and nonce
    return json.dumps({
        "ciphertext": encoder(ciphertext).decode('utf-8'),
        "tag": encoder(tag).decode('utf-8'),
        "nonce": encoder(nonce).decode('utf-8')
    })


def decrypt_from_json(key: Union[str, bytes], ciphertext_json: Union[str, bytes],
                      cipher: Union[SupportedCiphers, str] = SupportedCiphers.AES_GCM,
                      encoding: Union[SupportedEncodings, str] = SupportedEncodings.JSON_B64) -> str:
    """
    Decrypts data from a JSON string.

    Args:
        key (Union[str, bytes]): The decryption key, up to 16 bytes. Longer keys will be truncated.
        ciphertext_json (Union[str, bytes]): The encrypted data as a JSON string or bytes.
        cipher (SupportedEncodings): The cipher used for encryption.

    Returns:
        str: The decrypted plaintext data.

    Raises:
        InvalidCipher: If an unsupported cipher is provided.
        InvalidEncoding: If an unsupported encoding is provided.
        DecryptionKeyError: If decryption fails due to an invalid key or corrupted data.
    """
    cipher = _norm_cipher(cipher)
    encoding = _norm_encoding(encoding)

    if isinstance(ciphertext_json, str):
        ciphertext_json = json.loads(ciphertext_json)

    decoder = get_decoder(encoding)

    ciphertext: bytes = decoder(ciphertext_json["ciphertext"])

    if "tag" not in ciphertext_json:  # web crypto compatibility
        if cipher in AES_CIPHERS:
            ciphertext, tag = ciphertext[:-AES_TAG_SIZE], ciphertext[-AES_TAG_SIZE:]
        else:
            ciphertext, tag = ciphertext[:-CHACHA20_TAG_SIZE], ciphertext[-CHACHA20_TAG_SIZE:]
    else:
        tag = decoder(ciphertext_json["tag"])
    nonce = decoder(ciphertext_json["nonce"])

    try:
        ciphertext = decrypt_bin(key=key,
                                 ciphertext=nonce + ciphertext + tag,
                                 cipher=cipher)
        return ciphertext.decode("utf-8")
    except InvalidKeySize as e:
        raise e
    except Exception as e:
        raise DecryptionKeyError from e


def encrypt_bin(key: Union[str, bytes], plaintext: Union[str, bytes], cipher: Union[SupportedCiphers, str]) -> bytes:
    """
    Encrypts the given data and returns it as binary.

    Args:
        key (Union[str, bytes]): The encryption key, up to 16 bytes. Longer keys will be truncated.
        plaintext (Union[str, bytes]): The data to encrypt. Strings will be encoded as UTF-8.
        cipher (SupportedCiphers): The encryption cipher. Supported options:
            - AES_GCM: AES-GCM with 128-bit/256-bit key
            - CHACHA20_POLY1305: ChaCha20-Poly1305 with 256-bit key

    Returns:
        bytes: The encrypted data, including the nonce and tag.

    Raises:
        InvalidCipher: If an unsupported cipher is provided.
        InvalidKeySize: If an invalid key size is provided.
    """
    cipher = _norm_cipher(cipher)

    encryptor = encrypt_AES if cipher in AES_CIPHERS else encrypt_ChaCha20_Poly1305

    try:
        if cipher in BLOCK_CIPHERS:
            if cipher == SupportedCiphers.AES_GCM:
                mode = AES.MODE_GCM
            else:
                raise ValueError("invalid block cipher mode")
            ciphertext, tag, nonce = encryptor(key, plaintext, mode=mode)
        else:
            ciphertext, tag, nonce = encryptor(key, plaintext)
    except InvalidKeySize as e:
        raise e
    except Exception as e:
        raise EncryptionKeyError from e

    return nonce + ciphertext + tag


def decrypt_bin(key: Union[str, bytes], ciphertext: Union[str, bytes], cipher: Union[SupportedCiphers, str]) -> bytes:
    """
    Decrypts the given binary data.

    Args:
        key (Union[str, bytes]): The decryption key, up to 16 bytes. Longer keys will be truncated.
        ciphertext (Union[str, bytes]): The data to decrypt, including the nonce and tag.
        cipher (SupportedCiphers): The cipher used for encryption.

    Returns:
        bytes: The decrypted data.

    Raises:
        InvalidCipher: If an unsupported cipher is provided.
        DecryptionKeyError: If decryption fails due to an invalid key or corrupted data.
    """
    cipher = _norm_cipher(cipher)

    # extract nonce/tag depending on cipher, sizes are different
    if cipher == SupportedCiphers.AES_GCM:
        nonce, ciphertext, tag = (ciphertext[:AES_NONCE_SIZE],
                                  ciphertext[AES_NONCE_SIZE:-AES_TAG_SIZE],
                                  ciphertext[-AES_TAG_SIZE:])
    else:
        nonce, ciphertext, tag = (ciphertext[:CHACHA20_NONCE_SIZE],
                                  ciphertext[CHACHA20_NONCE_SIZE:-CHACHA20_TAG_SIZE],
                                  ciphertext[-CHACHA20_TAG_SIZE:])

    decryptor = decrypt_AES_128 if cipher in AES_CIPHERS else decrypt_ChaCha20_Poly1305
    try:
        if cipher in BLOCK_CIPHERS:
            if cipher == SupportedCiphers.AES_GCM:
                mode = AES.MODE_GCM
            else:
                raise ValueError("invalid block cipher mode")
            return decryptor(key, ciphertext, tag, nonce, mode=mode)
        return decryptor(key, ciphertext, tag, nonce)
    except InvalidKeySize as e:
        raise e
    except Exception as e:
        raise DecryptionKeyError from e


#############################
# Cipher Implementations
def encrypt_AES(key: Union[str, bytes], text: Union[str, bytes],
                nonce: Optional[bytes] = None,
                mode: Literal[AES.MODE_GCM] = AES.MODE_GCM) -> tuple[bytes, bytes, bytes]:
    """
    Encrypts plaintext using AES-GCM-128.

    Args:
        key (Union[str, bytes]): The encryption key. Strings will be encoded as UTF-8.
        text (Union[str, bytes]): The plaintext to encrypt.
        nonce (Optional[bytes]): An optional nonce. If None, a new one is generated.

    Returns:
        tuple[bytes, bytes, bytes]: The ciphertext, authentication tag, and nonce.
    """
    if not isinstance(text, bytes):
        text = bytes(text, encoding="utf-8")
    if not isinstance(key, bytes):
        key = bytes(key, encoding="utf-8")
    # AES-128 uses 128 bit/16 byte keys
    # AES-256 uses 256 bit/32 byte keys
    if len(key) not in AES_KEY_SIZES:
        raise InvalidKeySize("AES-GCM requires a 16/24/32 bytes key")
    cipher = AES.new(key, mode, nonce=nonce)
    ciphertext, tag = cipher.encrypt_and_digest(text)
    return ciphertext, tag, cipher.nonce


def decrypt_AES_128(key: Union[str, bytes],
                    ciphertext: bytes,
                    tag: bytes,
                    nonce: bytes,
                    mode: Literal[AES.MODE_GCM] = AES.MODE_GCM) -> bytes:
    """
    Decrypts ciphertext encrypted using AES-GCM-128.

    Args:
        key (Union[str, bytes]): The decryption key. Strings will be encoded as UTF-8.
        ciphertext (bytes): The encrypted ciphertext.
        tag (bytes): The authentication tag.
        nonce (bytes): The nonce used during encryption.

    Returns:
        str: The decrypted plaintext.

    Raises:
        InvalidKeySize: If key size is not valid
        ValueError: If decryption or authentication fails.
    """
    if isinstance(key, str):
        key = key.encode("utf-8")
    # AES-128 uses 128 bit/16 byte keys
    # AES-256 uses 256 bit/32 byte keys
    if len(key) not in AES_KEY_SIZES:
        raise InvalidKeySize("AES-GCM requires a 16/24/32 bytes key")
    cipher = AES.new(key, mode, nonce)
    return cipher.decrypt_and_verify(ciphertext, tag)


def encrypt_ChaCha20_Poly1305(key: Union[str, bytes],
                              text: Union[str, bytes],
                              nonce: Optional[bytes] = None) -> tuple[bytes, bytes, bytes]:
    """
    Encrypts plaintext using ChaCha20-Poly1305.

    Args:
        key (Union[str, bytes]): The encryption key. Strings will be encoded as UTF-8.
        text (Union[str, bytes]): The plaintext to encrypt.
        nonce (Optional[bytes]): An optional nonce. If None, a new one is generated.

    Returns:
        tuple[bytes, bytes, bytes]: The ciphertext, authentication tag, and nonce.
    """
    if isinstance(text, str):
        text = text.encode("utf-8")
    if isinstance(key, str):
        key = key.encode("utf-8")

    if len(key) != CHACHA20_KEY_SIZE:  # ChaCha20 uses 256 bit/32 byte keys
        raise InvalidKeySize("CHACHA20-POLY1305 requires a 32-byte key")
    if nonce:
        if len(nonce) != CHACHA20_NONCE_SIZE:  # 92bits/12bytes per RFC7539
            raise InvalidKeySize("CHACHA20-POLY1305 requires a 12-byte nonce per RFC7539")
    cipher = ChaCha20_Poly1305.new(key=key, nonce=nonce)
    ciphertext, tag = cipher.encrypt_and_digest(text)
    return ciphertext, tag, cipher.nonce


def decrypt_ChaCha20_Poly1305(key: Union[str, bytes],
                              ciphertext: bytes,
                              tag: bytes,
                              nonce: bytes) -> bytes:
    """
    Decrypts ciphertext encrypted using AES-GCM-128.

    Args:
        key (Union[str, bytes]): The decryption key. Strings will be encoded as UTF-8.
        ciphertext (bytes): The encrypted ciphertext.
        tag (bytes): The authentication tag.
        nonce (bytes): The nonce used during encryption.

    Returns:
        str: The decrypted plaintext.

    Raises:
        InvalidKeySize:
        ValueError: If decryption or authentication fails.
    """
    if isinstance(key, str):
        key = key.encode("utf-8")

    if len(key) != CHACHA20_KEY_SIZE:  # ChaCha20 uses 256 bit/32 byte keys
        raise InvalidKeySize("CHACHA20-POLY1305 requires a 32-byte key")
    if nonce:
        if len(nonce) != CHACHA20_NONCE_SIZE:  # 92bits/12bytes per RFC7539
            raise InvalidKeySize("CHACHA20-POLY1305 requires a 12-byte nonce per RFC7539")
    cipher = ChaCha20_Poly1305.new(key=key, nonce=nonce)
    return cipher.decrypt_and_verify(ciphertext, tag)


if __name__ == "__main__":
    from Cryptodome.Random import get_random_bytes

    print("JSON-B64" == SupportedEncodings.JSON_B64)

    key = get_random_bytes(CHACHA20_KEY_SIZE)
    plaintext = b'Attack at dawn'
    ciphertext, tag, nonce = encrypt_ChaCha20_Poly1305(key, plaintext)
    recovered = decrypt_ChaCha20_Poly1305(key, ciphertext, tag, nonce)
    print(recovered)
    assert recovered == plaintext




class UnsupportedProtocolVersion(ValueError):
    """ Specified protocol version is not supported """


class HiveMindException(Exception):
    """ An Exception inside the HiveMind"""


class UnauthorizedKeyError(HiveMindException):
    """ Invalid Key provided """


class InvalidCipher(HiveMindException):
    """unknown encryption scheme requested"""


class InvalidEncoding(HiveMindException):
    """unknown encoding scheme requested"""


class InvalidKeySize(HiveMindException):
    """ Encryption Key size does not obey specification"""


class WrongEncryptionKey(HiveMindException):
    """ Wrong Encryption Key"""


class DecryptionKeyError(WrongEncryptionKey):
    """ Could not decrypt payload """


class EncryptionKeyError(WrongEncryptionKey):
    """ Could not encrypt payload """


class HiveMindConnectionError(ConnectionError, HiveMindException):
    """ Could not connect to the HiveMind"""


class SecureConnectionFailed(HiveMindConnectionError):
    """ Could not connect by SSL """


class HiveMindEntryPointNotFound(HiveMindConnectionError):
    """ can not connect to provided address """


class Z85DecodeError(Exception):
    """Exception raised for errors in decoding Z85b."""
    pass



import pybase64
from dataclasses import dataclass
from typing import Optional, Tuple

from ovos_bus_client import Message as MycroftMessage
from ovos_bus_client import MessageBusClient
from ovos_bus_client.message import Message
from ovos_bus_client.session import Session, SessionManager
from ovos_utils.log import LOG

from hivemind_bus_client.client import HiveMessageBusClient
from hivemind_bus_client.encryption import SupportedEncodings, SupportedCiphers, optimal_ciphers
from hivemind_bus_client.identity import NodeIdentity
from hivemind_bus_client.message import HiveMessage, HiveMessageType
from poorman_handshake import HandShake, PasswordHandShake
from poorman_handshake.asymmetric.utils import decrypt_RSA, load_RSA_key


@dataclass()
class HiveMindSlaveInternalProtocol:
    """ this class handles all interactions between a hivemind listener and a ovos-core messagebus"""
    hm_bus: HiveMessageBusClient
    share_bus: bool = False
    bus: Optional[MessageBusClient] = None
    node_id: str = ""  # this is how ovos-core bus refers to this slave's master

    def register_bus_handlers(self):
        self.bus.on("hive.send.upstream", self.handle_send)
        self.bus.on("message", self.handle_outgoing_mycroft)  # catch all

    # mycroft handlers  - from slave -> master
    def handle_send(self, message: Message):
        """ ovos wants to send a HiveMessage

        a device can be both a master and a slave, upstream messages are handled here

        HiveMindListenerInternalProtocol will handle requests meant to go downstream
        """

        payload = message.data.get("payload")
        msg_type = message.data["msg_type"]

        hmessage = HiveMessage(msg_type, payload=payload)

        if msg_type == HiveMessageType.BROADCAST:
            # only masters can broadcast, ignore silently
            #   if this device is also a master to something,
            #   HiveMindListenerInternalProtocol will handle the request
            pass
        else:
            self.hm_bus.emit(hmessage)

    def handle_outgoing_mycroft(self, message: Message):
        """ forward internal messages to masters"""
        if isinstance(message, str):
            # "message" is a special case in ovos-bus-client that is not deserialized
            message = Message.deserialize(message)

        # this allows the master node to do passive monitoring of bus events
        if self.share_bus:
            msg = HiveMessage(HiveMessageType.SHARED_BUS,
                              payload=message.serialize())
            self.hm_bus.emit(msg)

        # this message is targeted at master
        # eg, a response to some bus event injected by master
        # note: master might completely ignore it
        peers = message.context.get("destination")
        if peers:
            if not isinstance(peers, list):
                peers = [peers]
            if self.node_id in peers:
                msg = HiveMessage(HiveMessageType.BUS,
                                  payload=message.serialize())
                self.hm_bus.emit(msg)


@dataclass()
class HiveMindSlaveProtocol:
    """
    Joins this instance ovos-core bus with master ovos-core bus
    Master becomes able to inject arbitrary bus messages
    """
    hm: HiveMessageBusClient
    identity: Optional[NodeIdentity] = None
    handshake: Optional[HandShake] = None
    pswd_handshake: Optional[PasswordHandShake] = None
    internal_protocol: HiveMindSlaveInternalProtocol = None
    mpubkey: str = ""  # asc public PGP key from master
    shared_bus: bool = False
    binarize: bool = False
    site_id: str = "unknown"

    def bind(self, bus: Optional[MessageBusClient] = None):
        if self.identity is None:
            self.identity = self.hm.identity or NodeIdentity()
        self.handshake = HandShake(self.identity.private_key)
        self.pswd_handshake = PasswordHandShake(self.identity.password) if self.identity.password else None

        if bus is None:
            bus = MessageBusClient()
            bus.run_in_thread()
            bus.connected_event.wait()
        LOG.info("Initializing HiveMindSlaveInternalProtocol")
        self.internal_protocol = HiveMindSlaveInternalProtocol(bus=bus, hm_bus=self.hm)
        self.internal_protocol.register_bus_handlers()
        LOG.info("registering protocol handlers")
        self.hm.on(HiveMessageType.HELLO, self.handle_hello)
        self.hm.on(HiveMessageType.BROADCAST, self.handle_broadcast)
        self.hm.on(HiveMessageType.PROPAGATE, self.handle_propagate)
        self.hm.on(HiveMessageType.INTERCOM, self.handle_intercom)
        self.hm.on(HiveMessageType.ESCALATE, self.handle_illegal_msg)
        self.hm.on(HiveMessageType.SHARED_BUS, self.handle_illegal_msg)
        self.hm.on(HiveMessageType.BUS, self.handle_bus)
        self.hm.on(HiveMessageType.HANDSHAKE, self.handle_handshake)

    @property
    def node_id(self):
        # this is how ovos-core bus refers to this slave's master
        return self.internal_protocol.node_id

    # hivemind events
    def handle_illegal_msg(self, message: HiveMessage):
        # this should not happen,
        # only sent from client -> server NOT server -> client
        # TODO log, kill connection (?)
        LOG.warning(f"illegal message {message}")

    def handle_hello(self, message: HiveMessage):
        # this check is because other nodes in the hive
        # may also send HELLO with their pubkey
        # only want this on the first connection
        LOG.info(f"HELLO: {message.payload}")
        if not self.node_id:
            self.mpubkey = message.payload.get("pubkey")
            node_id = message.payload.get("node_id", "")
            self.internal_protocol.node_id = node_id
            LOG.info(f"Connected to HiveMind: {node_id}")

    def start_handshake(self):
        if self.binarize:
            LOG.info("hivemind supports binarization protocol")
        else:
            LOG.info("hivemind does not support binarization protocol")

        payload = {"binarize": self.binarize,
                   "encodings": [SupportedEncodings.JSON_B64,
                                 SupportedEncodings.JSON_Z85B,
                                 SupportedEncodings.JSON_HEX],
                   "ciphers": optimal_ciphers()}
        if self.pswd_handshake is not None:
            payload["envelope"] = self.pswd_handshake.generate_handshake()
        else:
            payload["pubkey"] = self.handshake.pubkey

        self.hm.emit(HiveMessage(HiveMessageType.HANDSHAKE, payload))

    def receive_handshake(self, envelope):
        if self.pswd_handshake is not None:
            LOG.info("Received password envelope")
            self.pswd_handshake.receive_and_verify(envelope)  # validate master password matched
            self.hm.crypto_key = self.pswd_handshake.secret  # update to new crypto key
        else:
            LOG.info("Received pubkey envelope")
            # if we have a pubkey let's verify the master node is who it claims to be
            # currently this is sent in HELLO, but advance use cases can read it from somewhere else
            if self.mpubkey:
                # authenticates the server to the client
                self.handshake.receive_and_verify(envelope, self.mpubkey)
            else:
                # implicitly trust the server
                self.handshake.receive_handshake(envelope)
            self.hm.crypto_key = self.handshake.secret  # update to new crypto key

        # now that communication is secure, send our Session data and other personal info
        sess = Session(self.hm.session_id)
        msg = HiveMessage(HiveMessageType.HELLO, {"pubkey": self.identity.public_key,
                                                  "session": sess.serialize(),
                                                  "site_id": self.site_id})
        self.hm.emit(msg)
        self.hm.handshake_event.set()

    def handle_handshake(self, message: HiveMessage):
        LOG.info(f"HANDSHAKE: {message.payload}")
        # master is performing the handshake
        if "envelope" in message.payload:
            envelope = message.payload["envelope"]
            self.hm.json_encoding = message.payload.get("encoding") or SupportedEncodings.JSON_HEX
            self.hm.cipher = message.payload.get("cipher") or SupportedCiphers.AES_GCM
            self.receive_handshake(envelope)
            LOG.debug(f"Encoding: {self.hm.json_encoding}")
            LOG.debug(f"Cipher: {self.hm.cipher}")
            LOG.debug(f"Key size: {len(self.pswd_handshake.secret) * 8}bit")

        # master is requesting handshake start
        else:
            # required = message.payload.get("handshake")
            # if not required:
            #    self.hm.handshake_event.set()  # don't wait
            #    return

            encodings = message.payload.get("encodings") or [SupportedEncodings.JSON_HEX]
            ciphers = message.payload.get("ciphers") or [SupportedCiphers.AES_GCM]
            LOG.debug(f"Server supported encodings: {encodings}")
            LOG.debug(f"Server supported ciphers: {ciphers}")
            if message.payload.get("crypto_key") and self.hm.crypto_key:
                pass
                # we can use the pre-shared key instead of handshake
                # TODO - flag to give preference to pre-shared key over handshake

            self.binarize = message.payload.get("binarize", False)
            # TODO - flag to give preference to / require password or use RSA handshake
            # currently if password is set then it is always used
            if message.payload.get("password") and self.identity.password:
                self.pswd_handshake = PasswordHandShake(self.identity.password)
                self.start_handshake()

    def handle_bus(self, message: HiveMessage):
        LOG.info(f"BUS: {message.payload.msg_type}")
        assert isinstance(message.payload, MycroftMessage)
        # master wants to inject message into mycroft bus
        pload = message.payload

        # update session sent from hivemind-core
        sess = Session.from_message(pload)
        if sess.session_id == self.hm.session_id:
            sess.site_id = self.site_id  # do not allow overwriting site_id
        SessionManager.update(sess)

        # from this point on, it should be a native source and execute audio
        if "destination" in pload.context:
            pload.context["source"] = pload.context.pop("destination")
        self.internal_protocol.bus.emit(pload)

    def handle_broadcast(self, message: HiveMessage):
        LOG.info(f"BROADCAST: {message.payload}")

        if message.payload.msg_type == HiveMessageType.INTERCOM:
            if self.handle_intercom(message):
                return True

        if message.payload.msg_type == HiveMessageType.BUS:
            # if the message targets our site_id, send it to internal bus
            site = message.target_site_id
            if site and site == self.site_id:
                self.handle_bus(message.payload)

        # if this device is also a hivemind server
        # forward to HiveMindListenerInternalProtocol
        data = message.serialize()
        ctxt = {"source": self.node_id}
        self.internal_protocol.bus.emit(MycroftMessage('hive.send.downstream', data, ctxt))

    def handle_propagate(self, message: HiveMessage):
        LOG.info(f"PROPAGATE: {message.payload}")

        if message.payload.msg_type == HiveMessageType.INTERCOM:
            if self.handle_intercom(message):
                return True

        if message.payload.msg_type == HiveMessageType.BUS:
            # if the message targets our site_id, send it to internal bus
            site = message.target_site_id
            if site and site == self.site_id:
                # might originate from untrusted
                # satellite anywhere in the hive
                # do not inject by default
                pass  # TODO - when to inject ? add list of trusted peers?
                # self.handle_bus(message.payload)

        # if this device is also a hivemind server
        # forward to HiveMindListenerInternalProtocol
        data = message.serialize()
        ctxt = {"source": self.node_id}
        self.internal_protocol.bus.emit(MycroftMessage('hive.send.downstream', data, ctxt))

    def handle_intercom(self, message: HiveMessage):
        LOG.info(f"INTERCOM: {message.payload}")

        # if the message targets our site_id, send it to internal bus
        k = message.target_public_key
        if k and k != self.identity.public_key:
            # not for us
            return False

        pload = message.payload
        if isinstance(pload, dict) and "ciphertext" in pload:
            try:
                ciphertext = pybase64.b64decode(pload["ciphertext"])
                signature = pybase64.b64decode(pload["signature"])

                # TODO allow verifying, but we need to store known pubkeys before this is possible
                # pubkey = ""
                # verified: bool = verify_RSA(pubkey, ciphertext, signature)

                private_key = load_RSA_key(self.identity.private_key)

                decrypted = decrypt_RSA(private_key, ciphertext).decode("utf8")

                message._payload = HiveMessage.deserialize(decrypted)
            except:
                if k:
                    LOG.error("failed to decrypt message!")
                else:
                    LOG.debug("failed to decrypt message, not for us")
                return False

        if message.msg_type == HiveMessageType.BUS:
            self.handle_bus(message)
            return True
        elif message.msg_type == HiveMessageType.PROPAGATE:
            self.handle_propagate(message)
            return True
        elif message.msg_type == HiveMessageType.BROADCAST:
            self.handle_broadcast(message)
            return True
        return False
